import typing as t

from jinja2 import BaseLoader, ChoiceLoader, Environment, PackageLoader
from jinja2.ext import Extension
from pydantic import BaseSettings, Field, IPvAnyAddress, root_validator
from pydantic.color import Color


class DebugToolbarSettings(BaseSettings):
    DEFAULT_PANELS: t.List[str] = Field(
        [
            "debug_toolbar.panels.versions.VersionsPanel",
            "debug_toolbar.panels.timer.TimerPanel",
            "debug_toolbar.panels.settings.SettingsPanel",
            "debug_toolbar.panels.request.RequestPanel",
            "debug_toolbar.panels.headers.HeadersPanel",
            "debug_toolbar.panels.routes.RoutesPanel",
            "debug_toolbar.panels.logging.LoggingPanel",
            "debug_toolbar.panels.profiling.ProfilingPanel",
            "debug_toolbar.panels.redirects.RedirectsPanel",
        ],
        description=(
            "Specifies the full Python path to each panel that you "
            "want included in the toolbar."
        ),
    )
    PANELS: t.List[str] = Field(
        [],
        description=(
            "A list of the full Python paths to each panel that you "
            "want to append to `DEFAULT_PANELS`."
        ),
    )
    DISABLE_PANELS: t.Sequence[str] = Field(
        ["debug_toolbar.panels.redirects.RedirectsPanel"],
        description=(
            "A list of the full Python paths to each panel that you "
            "want disabled (but still displayed) by default."
        ),
    )
    ALLOWED_IPS: t.Optional[t.Sequence[IPvAnyAddress]] = Field(
        None,
        description=(
            "If it's set, the Debug Toolbar is shown only "
            "if your IP address is listed."
        ),
    )
    JINJA_ENV: Environment = Field(
        Environment(),
        description="The Jinja environment instance used to render the toolbar.",
    )
    JINJA_LOADERS: t.List[BaseLoader] = Field(
        [],
        description=(
            "Jinja `BaseLoader` subclasses used to load templates "
            "from the file system or other locations."
        ),
    )
    JINJA_EXTENSIONS: t.Sequence[t.Union[str, t.Type[Extension]]] = Field(
        [],
        description=(
            "Load the extensions from the list and bind them to the Jinja environment."
        ),
    )
    API_URL: str = Field(
        "/_debug_toolbar",
        description="URL prefix to use for toolbar endpoints.",
    )
    STATIC_URL: str = Field(
        f"{API_URL.default}/static",  # type: ignore
        description="URL to use when referring to toolbar static files.",
    )
    SHOW_TOOLBAR_CALLBACK: str = Field(
        "debug_toolbar.middleware.show_toolbar",
        description=(
            "This is the dotted path to a function used for "
            "determining whether the toolbar should show or not."
        ),
    )
    INSERT_BEFORE: str = Field(
        "</body>",
        description=(
            "The toolbar searches for this string in the HTML "
            "and inserts itself just before."
        ),
    )
    SHOW_COLLAPSE: bool = Field(
        False,
        description="If changed to `True`, the toolbar will be collapsed by default.",
    )
    ROOT_TAG_EXTRA_ATTRS: str = Field(
        "",
        description=(
            "This setting is injected in the root template div "
            "in order to avoid conflicts with client-side frameworks"
        ),
    )
    RESULTS_CACHE_SIZE: int = Field(
        25,
        description="The toolbar keeps up to this many results in memory.",
    )
    PROFILER_OPTIONS: t.Dict[str, t.Any] = Field(
        {"interval": 0.0001},
        description="A list of arguments can be supplied to the Profiler.",
    )
    SETTINGS: t.Sequence[BaseSettings] = Field(
        [],
        description=(
            "pydantic's `BaseSettings` instances to be "
            "displayed on the `SettingsPanel`."
        ),
    )
    LOGGING_COLORS: t.Dict[str, Color] = Field(
        {
            "CRITICAL": Color("rgba(255, 0, 0, .4)"),
            "ERROR": Color("rgba(255, 0, 0, .2)"),
            "WARNING": Color("rgba(255, 165, 0, .2)"),
            "INFO": Color("rgba(135, 206, 235, .2)"),
            "DEBUG": Color("rgba(128, 128, 128, .2)"),
        },
        description="Color palette used to apply colors based on the log level.",
    )
    SQL_WARNING_THRESHOLD: int = Field(
        500,
        description=(
            "The SQL panel highlights queries that took more that this amount of "
            "time, in milliseconds, to execute."
        ),
    )

    class Config:
        title = "Debug Toolbar"
        env_prefix = "DT_"
        case_sensitive = True

    def __init__(self, **settings: t.Any) -> None:
        super().__init__(**settings)
        loaders = self.JINJA_LOADERS + [PackageLoader("debug_toolbar", "templates")]
        self.JINJA_ENV.loader = ChoiceLoader(loaders)
        self.JINJA_ENV.trim_blocks = True
        self.JINJA_ENV.lstrip_blocks = True

        for extension in self.JINJA_EXTENSIONS:
            self.JINJA_ENV.add_extension(extension)

    @root_validator(pre=True)
    def ci(cls, values: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
        return {k.upper(): v for k, v in values.items()}



import typing as t

from fastapi import Request, Response

from debug_toolbar.panels import Panel


class RoutesPanel(Panel):
    title = "Routes"
    template = "panels/routes.html"

    async def generate_stats(
        self,
        request: Request,
        response: Response,
    ) -> t.Optional[t.Dict[str, t.Any]]:
        return {
            "routes": request.app.routes,
            "endpoint": request.scope["endpoint"],
        }



import typing as t

from fastapi import Request, Response
from starlette.middleware.base import RequestResponseEndpoint

from debug_toolbar.utils import get_name_from_obj

if t.TYPE_CHECKING:
    from debug_toolbar.toolbar import DebugToolbar


class Panel:
    has_content: bool = True

    def __init__(
        self,
        toolbar: "DebugToolbar",
        call_next: RequestResponseEndpoint,
    ) -> None:
        self.toolbar = toolbar
        self.call_next = call_next

    @property
    def panel_id(self) -> str:
        return self.__class__.__name__

    @property
    def enabled(self) -> bool:
        disabled_panels = self.toolbar.settings.DISABLE_PANELS
        panel_path = get_name_from_obj(self)

        disable_panel = (
            panel_path in disabled_panels
            or panel_path.replace(".panel.", ".") in disabled_panels
        )
        if disable_panel:
            default = "off"
        else:
            default = "on"

        return (
            self.toolbar.request.cookies.get(f"fastdt{self.panel_id}", default)
        ) == "on"

    @property
    def nav_title(self) -> str:
        return self.title

    @property
    def nav_subtitle(self) -> str:
        return ""

    @property
    def title(self) -> str:
        raise NotImplementedError

    @property
    def template(self) -> str:
        raise NotImplementedError

    @property
    def content(self) -> str:
        if self.has_content:
            return self.render(**self.get_stats())
        return ""

    def render(self, **context: t.Any) -> str:
        return self.toolbar.render(self.template, **context)

    def url_for(self, name: str, **path_params: t.Any) -> str:
        return self.toolbar.request.url_for(name, **path_params)

    @property
    def scripts(self) -> t.List[str]:
        return []

    async def process_request(self, request: Request) -> Response:
        return await self.call_next(request)

    async def generate_stats(
        self,
        request: Request,
        response: Response,
    ) -> t.Optional[t.Dict[str, t.Any]]:
        pass

    def get_stats(self) -> t.Dict[str, t.Any]:
        return self.toolbar.stats.get(self.panel_id, {})

    async def record_stats(self, request: Request, response: Response) -> None:
        stats = await self.generate_stats(request, response)

        if stats is not None:
            self.toolbar.stats.setdefault(self.panel_id, {}).update(stats)

    async def generate_server_timing(
        self,
        request: Request,
        response: Response,
    ) -> t.Optional[t.Sequence[t.Sequence[t.Union[str, float]]]]:
        pass

    def get_server_timing_stats(self) -> t.Sequence[t.Sequence[t.Union[str, float]]]:
        return self.toolbar.server_timing_stats.get(self.panel_id, [])

    async def record_server_timing(self, request: Request, response: Response) -> None:
        stats = await self.generate_server_timing(request, response)

        if stats is not None:
            st_stats = self.toolbar.server_timing_stats.setdefault(self.panel_id, [])
            st_stats += list(stats)



import typing as t
from collections import defaultdict
from time import perf_counter

from fastapi import Request, Response
from fastapi.dependencies.utils import solve_dependencies
from fastapi.routing import APIRoute
from sqlalchemy import event
from sqlalchemy.engine import Connection, Engine
from sqlalchemy.engine.default import DefaultExecutionContext
from sqlalchemy.orm import Session

from debug_toolbar.panels import Panel
from debug_toolbar.sql import parse_sql, simplify
from debug_toolbar.utils import color_generator, matched_route


class SQLAlchemyPanel(Panel):
    title = "SQLAlchemy"
    template = "panels/sqlalchemy.html"

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self._sql_time: int = 0
        self._queries: t.List[t.Tuple[str, t.Dict[str, t.Any]]] = []
        self._databases: t.Dict[str, t.Dict[str, t.Any]] = {}
        self._colors: t.Generator[t.Tuple[int, ...], None, None] = color_generator()

    @property
    def nav_subtitle(self) -> str:
        count = len(self._queries)
        return (
            f"{count} {'query' if count == 1 else 'queries'}"
            f" in {self._sql_time:.2f}ms"
        )

    def register(self, engine: Engine) -> None:
        event.listen(engine, "before_cursor_execute", self.before_cursor_execute)
        event.listen(engine, "after_cursor_execute", self.after_cursor_execute)

    def before_cursor_execute(
        self,
        conn: Connection,
        cursor: t.Any,
        statement: str,
        parameters: t.Union[t.Sequence, t.Dict],
        context: DefaultExecutionContext,
        executemany: bool,
    ) -> None:
        context._start_time = perf_counter()

    def after_cursor_execute(
        self,
        conn: Connection,
        cursor: t.Any,
        statement: str,
        parameters: t.Union[t.Sequence, t.Dict],
        context: DefaultExecutionContext,
        executemany: bool,
    ) -> None:
        end_time = perf_counter()
        duration = (end_time - context._start_time) * 1000
        compiled = context.invoked_statement.compile(
            compile_kwargs={"literal_binds": True}
        )
        sql = str(compiled)
        alias = str(conn.engine.url)

        query = {
            "statement": statement,
            "params": parameters,
            "start_time": context._start_time,
            "end_time": end_time,
            "duration": duration,
            "sql_raw": sql,
            "sql_formatted": parse_sql(sql, aligned_indent=True),
            "sql_simple": simplify(parse_sql(sql, aligned_indent=False)),
            "is_slow": duration > self.toolbar.settings.SQL_WARNING_THRESHOLD,
            "is_select": sql.lower().strip().startswith("select"),
        }
        if alias not in self._databases:
            self._databases[alias] = {
                "time_spent": duration,
                "num_queries": 1,
                "rgb_color": next(self._colors),
            }
        else:
            self._databases[alias]["time_spent"] += duration
            self._databases[alias]["num_queries"] += 1

        self._sql_time += duration
        self._queries.append((alias, query))

    async def process_request(self, request: Request) -> Response:
        route = matched_route(request)

        if hasattr(route, "dependant"):
            route = t.cast(APIRoute, route)

            solved_result = await solve_dependencies(
                request=request,
                dependant=route.dependant,
                dependency_overrides_provider=route.dependency_overrides_provider,
            )
            for value in solved_result[0].values():
                if isinstance(value, Session):
                    self.register(value.get_bind())

        return await super().process_request(request)

    async def generate_stats(
        self,
        request: Request,
        response: Response,
    ) -> t.Optional[t.Dict[str, t.Any]]:
        trace_colors: t.Dict[str, t.Tuple[int, ...]] = defaultdict(
            lambda: next(self._colors)
        )
        query_dups: t.Dict[str, t.Dict[str, int]] = defaultdict(
            lambda: defaultdict(int)
        )
        query_similar: t.Dict[str, t.Dict[str, t.Any]] = defaultdict(
            lambda: defaultdict(int)
        )
        width_ratio_tally = 0

        for alias, query in self._queries:
            query_dups[alias][query["sql_raw"]] += 1
            query_similar[alias][query["statement"]] += 1
            try:
                width_ratio = (query["duration"] / self._sql_time) * 100
            except ZeroDivisionError:
                width_ratio = 0

            query.update(
                {
                    "trace_color": trace_colors[query["sql_raw"]],
                    "start_offset": width_ratio_tally,
                    "end_offset": width_ratio + width_ratio_tally,
                    "width_ratio": width_ratio,
                }
            )
            width_ratio_tally += width_ratio

        query_dups = {
            alias: {
                query: dup_count
                for query, dup_count in queries.items()
                if dup_count >= 2
            }
            for alias, queries in query_dups.items()
        }
        query_similar = {
            alias: {
                query: (similar_count, next(self._colors))
                for query, similar_count in queries.items()
                if similar_count >= 2
            }
            for alias, queries in query_similar.items()
        }
        for alias, query in self._queries:
            try:
                (query["similar_count"], query["similar_color"]) = query_similar[alias][
                    query["statement"]
                ]
                query["dup_count"] = query_dups[alias][query["sql_raw"]]
            except KeyError:
                continue

        for alias, info in self._databases.items():
            try:
                info["similar_count"] = sum(c for c, _ in query_similar[alias].values())
                info["dup_count"] = sum(c for c in query_dups[alias].values())
            except KeyError:
                continue

        return {
            "databases": self._databases,
            "queries": self._queries,
        }



import asyncio
import functools
import inspect
import sys
import typing as t

from fastapi import Request
from fastapi.routing import APIRoute
from starlette.routing import Match
from starlette.staticfiles import StaticFiles


def import_string(import_name: str) -> t.Any:
    try:
        __import__(import_name)
    except ImportError:
        if "." not in import_name:
            raise
    else:
        return sys.modules[import_name]

    module_name, obj_name = import_name.rsplit(".", 1)
    module = __import__(module_name, globals(), locals(), [obj_name])
    try:
        return getattr(module, obj_name)
    except AttributeError as e:
        raise ImportError(e)


def get_name_from_obj(obj: t.Any) -> str:
    if hasattr(obj, "__name__"):
        name = obj.__name__
    else:
        name = obj.__class__.__name__

    if hasattr(obj, "__module__"):
        module = obj.__module__
        name = f"{module}.{name}"
    return name


def matched_route(request: Request) -> t.Optional[APIRoute]:
    for route in request.app.routes:
        match, _ = route.matches(request.scope)
        if match == Match.FULL:
            return route
    return None


def matched_endpoint(request: Request) -> t.Optional[t.Callable]:
    route = matched_route(request)
    if route is not None:
        endpoint = getattr(route, "endpoint", None)

        if endpoint is not None:
            return endpoint
        if not isinstance(route.app, StaticFiles):
            return route.app
    return None


def is_coroutine(endpoint: t.Callable) -> bool:
    handler = endpoint

    while isinstance(handler, functools.partial):
        handler = handler.func
    if not (inspect.ismethod(handler) or inspect.isfunction(handler)):
        handler = handler.__call__  # type: ignore
    return asyncio.iscoroutinefunction(handler)


def pluralize(value: float, arg: str = "s") -> str:
    if "," not in arg:
        arg = f",{arg}"

    bits = arg.split(",")

    if len(bits) > 2:
        return ""

    singular_suffix, plural_suffix = bits[:2]

    if float(value) == 1:
        return singular_suffix
    return plural_suffix


def color_generator() -> t.Generator[t.Tuple[int, ...], None, None]:
    triples = [
        (0, 1, 1),
        (1, 1, 0),
        (1, 0, 1),
        (1, 1, 1),
        (0, 1, 0),
        (0, 0, 1),
    ]
    n = 1 << 7
    so_far = [[0, 0, 0]]
    while True:
        if n == 0:
            yield (0, 0, 0)
        copy_so_far = list(so_far)
        for triple in triples:
            for previous in copy_so_far:
                rgb = [n * triple[i] + previous[i] for i in range(3)]
                so_far.append(rgb)
                yield tuple(rgb)
        n >>= 1



import typing as t

from fastapi import Request, Response

from debug_toolbar.panels import Panel
from debug_toolbar.utils import get_name_from_obj


class RequestPanel(Panel):
    title = "Request"
    template = "panels/request.html"

    @property
    def nav_subtitle(self) -> str:
        return get_name_from_obj(self.endpoint)

    async def generate_stats(
        self,
        request: Request,
        response: Response,
    ) -> t.Optional[t.Dict[str, t.Any]]:
        self.endpoint = request.scope["endpoint"]
        stats: t.Dict[str, t.Any] = {"request": request}

        if hasattr(self, "_form"):
            stats["form"] = await request.form()

        if "session" in request.scope:
            stats["session"] = request.session.items()
        return stats



import html
import re
import typing as t

import sqlparse
from sqlparse import tokens as T


class BoldKeywordFilter:
    def process(
        self,
        stream: t.Generator[t.Tuple[T._TokenType, str], None, None],
    ) -> t.Generator[t.Tuple[t.Type[T.Text], str], None, None]:
        for token_type, value in stream:
            is_keyword = token_type in T.Keyword
            if is_keyword:
                yield T.Text, "<strong>"
            yield token_type, html.escape(value)
            if is_keyword:
                yield T.Text, "</strong>"


def simplify(sql: str) -> str:
    expr = r"SELECT</strong> (...........*?) <strong>FROM"
    sub = r"SELECT</strong> &#8226;&#8226;&#8226; <strong>FROM"
    return re.sub(expr, sub, sql)


def parse_sql(sql: str, aligned_indent: bool = False) -> str:
    stack = sqlparse.engine.FilterStack()
    stack.enable_grouping()

    if aligned_indent:
        stack.stmtprocess.append(
            sqlparse.filters.AlignedIndentFilter(char="&nbsp;", n="<br/>")
        )
    stack.preprocess.append(BoldKeywordFilter())
    stack.postprocess.append(sqlparse.filters.SerializerUnicode())
    return "".join(stack.run(sql))



import typing as t

from fastapi import Request, Response

from debug_toolbar.panels import Panel


class HeadersPanel(Panel):
    title = "Headers"
    template = "panels/headers.html"

    async def generate_stats(
        self,
        request: Request,
        response: Response,
    ) -> t.Optional[t.Dict[str, t.Any]]:
        request_headers = dict(request.headers)

        if "cookie" in request_headers:
            request_headers["cookie"] = "=> see Request panel"

        return {
            "request_headers": request_headers,
            "environ": request.scope.get("asgi", {}),
            "response_headers": response.headers,
        }



import html
import typing as t

from fastapi import Request

from debug_toolbar.toolbar import DebugToolbar


def render_panel(request: Request, store_id: str, panel_id: str) -> t.Any:
    toolbar = DebugToolbar.fetch(store_id)

    if toolbar is None:
        content = (
            "Data for this panel isn't available anymore. "
            "Please reload the page and retry."
        )
        content = f"<p>{html.escape(content)}</p>"
        scripts = []
    else:
        panel = toolbar.get_panel_by_id(panel_id)
        content = panel.content
        scripts = panel.scripts

    return {"content": content, "scripts": scripts}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['debug_toolbar', 'debug_toolbar.panels']

package_data = \
{'': ['*'],
 'debug_toolbar': ['statics/css/*',
                   'statics/img/*',
                   'statics/js/*',
                   'templates/*',
                   'templates/includes/*',
                   'templates/panels/*']}

install_requires = \
['Jinja2>=2.9',
 'aiofiles>=0.2.1',
 'fastapi>=0.62.0',
 'pyinstrument>=3.0.0',
 'sqlparse>=0.2.0']

setup_kwargs = {
    'name': 'fastapi-debug-toolbar',
    'version': '0.1.1',
    'description': 'A debug toolbar for FastAPI.',
    'long_description': '# ![FastAPI](https://raw.githubusercontent.com/mongkok/fastapi-debug-toolbar/main/debug_toolbar/statics/img/icon-green.svg) Debug Toolbar\n\n<p align="center">\n    <a href="https://fastapi-debug-toolbar.domake.io">\n        <img src="https://user-images.githubusercontent.com/5514990/127749737-1a2b5ce7-a191-4f8b-96e4-fe94e0a08f6f.gif" alt="FastAPI Debug Toolbar">\n    </a>\n</p>\n<p align="center">\n    <em>🐞A debug toolbar for FastAPI based on the original django-debug-toolbar.🐞</em>\n    <br><em><b>Swagger UI</b> & <b>GraphQL</b> are supported.</em>\n</p>\n<p align="center">\n<a href="https://github.com/mongkok/fastapi-debug-toolbar/actions">\n    <img src="https://github.com/mongkok/fastapi-debug-toolbar/actions/workflows/test-suite.yml/badge.svg" alt="Test">\n</a>\n<a href="https://codecov.io/gh/mongkok/fastapi-debug-toolbar">\n    <img src="https://img.shields.io/codecov/c/github/mongkok/fastapi-debug-toolbar?color=%2334D058" alt="Coverage">\n</a>\n<a href="https://www.codacy.com/gh/mongkok/fastapi-debug-toolbar/dashboard">\n    <img src="https://app.codacy.com/project/badge/Grade/e9d8ba3973264424a3296016063b4ab5" alt="Codacy">\n</a>\n<a href="https://pypi.org/project/fastapi-debug-toolbar">\n    <img src="https://img.shields.io/pypi/v/fastapi-debug-toolbar" alt="Package version">\n</a>\n\n---\n\n**Documentation**: [https://fastapi-debug-toolbar.domake.io](https://fastapi-debug-toolbar.domake.io)\n\n---\n\n## Installation\n\n```sh\npip install fastapi-debug-toolbar\n```\n\n## Quickstart\n\nAdd `DebugToolbarMiddleware` middleware to your FastAPI application:\n\n```py\nfrom debug_toolbar.middleware import DebugToolbarMiddleware\nfrom fastapi import FastAPI\n\napp = FastAPI(debug=True)\napp.add_middleware(DebugToolbarMiddleware)\n```\n\n## SQLAlchemy\n\nPlease make sure to use the *"Dependency Injection"* system as described in the [FastAPI docs](https://fastapi.tiangolo.com/tutorial/sql-databases/#create-a-dependency) and add the SQLAlchemy panel to the argument list of the Debug Toolbar middleware:\n\n```py\napp.add_middleware(\n    DebugToolbarMiddleware,\n    panels=["debug_toolbar.panels.sqlalchemy.SQLAlchemyPanel"],\n)\n```\n',
    'author': 'Dani',
    'author_email': 'dani@domake.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mongkok/fastapi-debug-toolbar',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import time
import typing as t

from fastapi import Request, Response

from debug_toolbar.panels import Panel

try:
    import resource
except ImportError:
    resource = None  # type: ignore


class TimerPanel(Panel):
    has_content = resource is not None
    title = "Time"
    template = "panels/timer.html"

    @property
    def nav_subtitle(self) -> str:
        stats = self.get_stats()

        if hasattr(self, "_start_ru"):
            utime = self._end_ru.ru_utime - self._start_ru.ru_utime
            stime = self._end_ru.ru_stime - self._start_ru.ru_stime
            return f"CPU: {(utime + stime) * 1000.0:.2f}ms ({stats['elapsed']:.2f}ms)"

        return f"Total: {stats['elapsed']:.2f}ms"

    @property
    def content(self) -> str:
        stats = self.get_stats()

        rows = (
            ("User CPU time", f"{stats['utime']:.3f} msec"),
            ("System CPU time", f"{stats['stime']:.3f} msec"),
            ("Total CPU time", f"{stats['total']:.3f} msec"),
            ("Elapsed time", f"{stats['elapsed']:.3f} msec"),
            (
                "Context switches",
                f"{stats['vcsw']} voluntary, {stats['ivcsw']} involuntary",
            ),
        )
        return self.render(rows=rows)

    @property
    def scripts(self) -> t.List[str]:
        scripts = super().scripts
        scripts.append(self.url_for("debug_toolbar.static", path="js/timer.js"))
        return scripts

    async def process_request(self, request: Request) -> Response:
        self._start_time = time.time()
        if self.has_content:
            self._start_ru = resource.getrusage(resource.RUSAGE_SELF)
        return await super().process_request(request)

    async def generate_stats(
        self,
        request: Request,
        response: Response,
    ) -> t.Optional[t.Dict[str, t.Any]]:
        stats = {
            "elapsed": (time.time() - self._start_time) * 1000,
        }
        if hasattr(self, "_start_ru"):
            self._end_ru = resource.getrusage(resource.RUSAGE_SELF)
            utime = 1000 * self._elapsed_ru("ru_utime")
            stime = 1000 * self._elapsed_ru("ru_stime")

            stats.update(
                {
                    "utime": utime,
                    "stime": stime,
                    "total": utime + stime,
                    "vcsw": self._elapsed_ru("ru_nvcsw"),
                    "ivcsw": self._elapsed_ru("ru_nivcsw"),
                }
            )
        return stats

    async def generate_server_timing(
        self,
        request: Request,
        response: Response,
    ) -> t.Optional[t.Sequence[t.Sequence[t.Union[str, float]]]]:
        stats = self.get_stats()

        return [
            ("utime", "User CPU time", stats.get("utime", 0)),
            ("stime", "System CPU time", stats.get("stime", 0)),
            ("total", "Total CPU time", stats.get("total", 0)),
            ("elapsed", "Elapsed time", stats.get("elapsed", 0)),
        ]

    def _elapsed_ru(self, name: str) -> float:
        return getattr(self._end_ru, name) - getattr(self._start_ru, name)



import datetime
import logging
import typing as t

from fastapi import Request, Response
from starlette.concurrency import run_in_threadpool

from debug_toolbar.panels import Panel
from debug_toolbar.utils import is_coroutine, matched_endpoint, pluralize

try:
    import threading
except ImportError:
    threading = None  # type: ignore


class LogCollector:
    def __init__(self) -> None:
        if threading is None:
            raise NotImplementedError(
                "threading module is not available, "
                "this panel cannot be used without it"
            )
        self.collections: t.Dict[t.Any, t.List[t.Dict[str, t.Any]]] = {}

    def get_collection(self, thread: t.Any = None) -> t.List[t.Dict[str, t.Any]]:
        if thread is None:
            thread = threading.current_thread()
        if thread not in self.collections:
            self.collections[thread] = []
        return self.collections[thread]

    def clear_collection(self, thread: t.Any = None) -> None:
        if thread is None:
            thread = threading.current_thread()
        if thread in self.collections:
            del self.collections[thread]

    def collect(self, item: t.Dict[str, t.Any], thread: t.Any = None) -> None:
        self.get_collection(thread).append(item)


class ThreadTrackingHandler(logging.Handler):
    def __init__(self, collector: LogCollector) -> None:
        logging.Handler.__init__(self)
        self.collector = collector

    def emit(self, record: logging.LogRecord) -> None:
        try:
            message = record.getMessage()
        except Exception:
            message = "[Could not get log message]"

        self.collector.collect(
            {
                "message": message,
                "time": datetime.datetime.fromtimestamp(record.created),
                "level": record.levelname,
                "file": record.pathname,
                "line": record.lineno,
                "channel": record.name,
            }
        )


collector = LogCollector()
logging_handler = ThreadTrackingHandler(collector)
logging.root.addHandler(logging_handler)


class LoggingPanel(Panel):
    nav_title = "Logging"
    title = "Log messages"
    template = "panels/logging.html"

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self._records: t.Dict[t.Any, t.List[t.Dict[str, t.Any]]] = {}

    @property
    def nav_subtitle(self) -> str:
        stats = self.get_stats()
        record_count = len(stats["records"])
        return f"{record_count} message{pluralize(record_count)}"

    async def process_request(self, request: Request) -> Response:
        endpoint = matched_endpoint(request)

        if endpoint is None:
            return await super().process_request(request)

        if is_coroutine(endpoint):
            self.current_thread = threading.current_thread()
        else:
            self.current_thread = await run_in_threadpool(threading.current_thread)

        collector.clear_collection(thread=self.current_thread)
        return await super().process_request(request)

    async def generate_stats(
        self,
        request: Request,
        response: Response,
    ) -> t.Optional[t.Dict[str, t.Any]]:
        records = collector.get_collection(thread=self.current_thread)
        self._records[self.current_thread] = records
        collector.clear_collection(thread=self.current_thread)
        return {"records": records}
